@@ -3,7 +3,10 @@ import re
 import os
 import sys
 import json
+import uuid
 import pickle
+
+from numpy import insert
 import humanize
 import datetime
 import inspect
@@ -12,13 +15,13 @@ from time import time
 from functools import wraps
 from datetime import datetime
 from typing import Tuple, Callable, Union, Dict, List, Optional
-from .utils import get_params, parse_time, random_string
+from .utils import get_params, parse_time, random_string, get_format_groups, get_next_format_group
 from .backends import Backend
 
 # Dictionary are not hashable and the python hash is not consistent
 # between runs so we have to use an external dictionary hashing package
 # else we will not be able to load the saved caches.
-from dict_hash import sha256
+from dict_hash import sha256, Hashable
 
 log_levels = {
     "debug":logging.DEBUG,
@@ -421,19 +424,13 @@ class Cache:
         # actually load the values
         return Backend(self.load_kwargs, self.dump_kwargs).load(metadata.get("backend_metadata", {}), path)
 
-    def _dump(self, args, kwargs, result, path, start_time, end_time):
-
+    def _check_return_type_compatability(self, result, path):
         # Check if it's a structured path
         if isinstance(path, list) or isinstance(path, tuple):
             assert isinstance(result,list) or isinstance(result,tuple)
             assert len(result) == len(path)
-            for r, p in zip(result, path):
-                self._dump(args, kwargs, r, p, start_time, end_time)
-            return 
-                
         elif isinstance(path, dict):
             assert isinstance(result, dict)
-
             required_keys = set(path.keys()) - set(self.optional_path_keys)
             missing_keys = required_keys - set(result.keys())
             if len(missing_keys) != 0:
@@ -449,7 +446,15 @@ class Cache:
                     "not appear in the defined path. In particular the the "
                     "extra keys are '{}'"
                 ).format(result.keys(), extra_keys))
+            return 
 
+    def _dump(self, args, kwargs, result, path, start_time, end_time):
+        # Check if it's a structured path
+        if isinstance(path, list) or isinstance(path, tuple):
+            for r, p in zip(result, path):
+                self._dump(args, kwargs, r, p, start_time, end_time)
+            return 
+        elif isinstance(path, dict):
             for key in result.keys():
                 self._dump(args, kwargs, result[key], path[key], start_time, end_time)
             return 
@@ -525,18 +530,22 @@ class Cache:
             json.dump(metadata, f, indent=4)
 
 
-    def _decorate_callable(self, function: Callable) -> Callable:
+    def _decorate_function(self, function: Callable) -> Callable:
         # wraps to support pickling
         @wraps(function)
         def wrapped(*args, **kwargs):
             cache_enabled, args, kwargs = self._is_cache_enabled(args, kwargs)
+            
             # if the cache is not enabled just forward the call
             if not cache_enabled:
                 self.logger.info("The cache is disabled")
-                return function(*args, **kwargs)
+                result = function(*args, **kwargs)
+                self._check_return_type_compatability(result, self.cache_path)
+                return result
 
             # Get the path
             path = self._get_formatted_path(args, kwargs)
+
             # Try to load the cache
             result = self._load(path)
             # if we got a result, reutrn it
@@ -551,6 +560,7 @@ class Cache:
 
             # Save the result
             try:
+                self._check_return_type_compatability(result, path)
                 self._dump(args, kwargs, result, path, start_time, end_time)
             except Exception as e:
                 if self.is_backup_enabled:
@@ -565,7 +575,56 @@ class Cache:
         setattr(wrapped, "__cacher_instance", self)
         return wrapped
 
-    def _get_formatted_path(self, args, kwargs, formatter=None, function_info=None, extra_kwargs=None) -> str:
+    def _decorate_method(self, function: Callable) -> Callable:
+        # wraps to support pickling
+        @wraps(function)
+        def wrapped(self, *args, **kwargs):
+            cache_enabled, args, kwargs = self._is_cache_enabled(args, kwargs)
+            
+            # if the cache is not enabled just forward the call
+            if not cache_enabled:
+                self.logger.info("The cache is disabled")
+                result = function(self, *args, **kwargs)
+                self._check_return_type_compatability(result, self.cache_path)
+                return result
+
+            # Get the path
+            path = self._get_formatted_path(args, kwargs, inner_self=self)
+
+            # Check that the self is actually hashable
+            if not issubclass(self, Hashable):
+                raise ValueError("Could not has self of class `{}` because it doesn't implement Hashable (from dict_hash).".format(self.__class__.__name__))
+
+            # Try to load the cache
+            result = self._load(path)
+            # if we got a result, reutrn it
+            if result is not None:
+                return result
+            
+            self.logger.info("Computing the result for %s %s", args, kwargs)
+            # otherwise compute the result
+            start_time = time()
+            result = function(*args, **kwargs)
+            end_time = time()
+
+            # Save the result
+            try:
+                self._check_return_type_compatability(result, path)
+                self._dump(args, kwargs, result, path, start_time, end_time)
+            except Exception as e:
+                if self.is_backup_enabled:
+                    raise self._backup(result, path, e, args, kwargs)
+                raise e
+
+            return result
+
+        # add a reference to the cached function so we can unpack
+        # The caching if needed
+        setattr(wrapped, "__cached_function", function)
+        setattr(wrapped, "__cacher_instance", self)
+        return wrapped
+
+    def _get_formatted_path(self, args, kwargs, formatter=None, function_info=None, extra_kwargs=None, inner_self=None) -> str:
         """Compute the path adding and computing the needed arguments."""        
         formatter = formatter or self.cache_path
 
@@ -591,35 +650,65 @@ class Cache:
 
         function_info = function_info or self.function_info
         params = get_params(function_info, args, kwargs)
-        
-        if "_hash" in formatter:
-            params["_hash"] = sha256({"params": params, "function_info": function_info})
+        groups = get_format_groups(formatter)
+
+        if "_hash" in {match.str_match for match in groups}:
+            data = {"params": params, "function_info": function_info}
+
+            if inner_self is not None: 
+                data["self"] = inner_self
+
+            params["_hash"] = sha256(data)
 
         self.logger.debug("Got parameters %s", params)
 
-        # Handle the composite paths
-        for match in re.finditer(r"\{([^\{]+)(:?\..+?)+\}", formatter):
-            # Extract the matching string
-            match = match.group(0)
-            # Get the name of the base element and the attributes chain
-            root, *attrs = match[1:-1].split(".")
-            # Get the params to use for the attributes chain
-            root = params[root]
-            # Follow the attributes chain
-            for attr in attrs:
-                root = getattr(root, attr)
-            # Replace the result in the formatter
-            formatter = formatter.replace(match, str(root))
-
-        # Compute the path of the cache for these parameters
-        path = formatter.format(
-            cache_dir=self.cache_dir,
+        format_args = {
             **params,
             **function_info,
             **extra_kwargs,
+            "cache_dir":self.cache_dir,
+        }
+
+        new_formatter = ""
+        old_formatter = formatter
+        # Handle the composite paths
+        while len(old_formatter) != 0:
+            new_match, formatter_remainder = get_next_format_group(old_formatter)
+
+            # there are no more matches just append the remainder
+            if new_match is None:
+                new_formatter += formatter_remainder
+                break
+            
+            # check if we should call the value or not
+            if new_match.str_match.endswith("()"):
+                new_match.str_match  = new_match.str_match[:-2]
+                # Get the name of the base element and the attributes chain
+                root, *attrs = new_match.str_match.split(".")
+                # Get the params to use for the attributes chain
+                root = format_args[root]
+
+                # Follow the attributes chain
+                for attr in attrs:                    
+                    root = getattr(root, attr)
+
+                # Check if we have to call the function or not
+                if inspect.isfunction(root) or inspect.ismethod(root):
+                    root = root()
+
+                sub = str(root)
+            else:
+                sub = "{" + new_match.str_match + "}"
+
+            new_formatter += old_formatter[:new_match.start]
+            new_formatter += sub
+            old_formatter = formatter_remainder
+
+        path = new_formatter.format(
+            **format_args,
         )
         self.logger.debug("Calculated path %s", path)
-
+        print(path)
         return path
 
     def _fix_docs(self, function: Callable, wrapped: Callable) -> Callable:
@@ -632,7 +721,12 @@ class Cache:
     def decorate(self, function: Callable) -> Callable:
         self.function_info = self._compute_function_info(function)
         self.decorated_function = function
-        wrapped = self._decorate_callable(function)
+
+        if inspect.ismethod(function):
+            wrapped = self._decorate_method(function)
+        else:
+            wrapped = self._decorate_function(function)
+
         wrapped = self._fix_docs(function, wrapped)
         return wrapped
 
